/*
 * $Id: $
 * $Date: $
 *
 * Author:	Kenneth Chin-Purcell, AHPCRC
 *		Copyright @ 1992, Minnesota Supercomputer Center, Inc.
 *
 * Purpose:	magnifier
 *
 * RESTRICTED RIGHTS LEGEND
 *
 * Use, duplication, or disclosure of this software and its documentation
 * by the Government is subject to restrictions as set forth in subdivision
 * { (b) (3) (ii) } of the Rights in Technical Data and Computer Software
 * clause at 52.227-7013.
 *
 * $Log:$
 */

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xproto.h>		/* To catch error on TranslateCoord */
#include "PuffP.h"
#include "util.h"

extern int atoi(char *);

/* Resources
 */
#define CoreOff(field) XtOffset(Widget, core.field)
#define PuffOff(field) XtOffset(PuffWidget, puff.field)

static Dimension	widthDim = 500;
static Dimension	heightDim = 300;

static XtResource resources[] = {
    { XtNwidth, 	XtCWidth, 	XtRDimension, sizeof(Dimension),
	  CoreOff(width),	XtRDimension, (XtPointer) &widthDim },
    { XtNheight, 	XtCHeight, 	XtRDimension, sizeof(Dimension),
	  CoreOff(height),	XtRDimension, (XtPointer) &heightDim },
    { XtNmagX,		XtCMag,		XtRInt, sizeof(int),
	  PuffOff(magx),	XtRImmediate, (XtPointer) 6 },
    { XtNmagY,		XtCMag,		XtRInt, sizeof(int),
	  PuffOff(magy),	XtRImmediate, (XtPointer) 6 },
    { XtNsrcX,		XtCSrcX,	XtRInt, sizeof(int),
	  PuffOff(srcx),	XtRImmediate, (XtPointer) 6 },
    { XtNsrcY,		XtCSrcY,	XtRInt, sizeof(int),
	  PuffOff(srcy),	XtRImmediate, (XtPointer) 6 },
    { XtNcursorWidth,	XtCCursorWidth,	XtRInt, sizeof(int),
	  PuffOff(cursorSize),	XtRImmediate, (XtPointer) 3 },
    { XtNdelayMin,	XtCDelayMin,	XtRInt, sizeof(int),
	  PuffOff(delayMin),	XtRImmediate, (XtPointer) 50 },
    { XtNdelayMax,	XtCDelayMax,	XtRInt, sizeof(int),
	  PuffOff(delayMax),	XtRImmediate, (XtPointer) 500 },
    { XtNviewBox,	XtCViewBox,	XtRInt, sizeof(int),
	  PuffOff(viewBox),	XtRImmediate, (XtPointer) 1 },
};


/* Methods
 */
static void Initialize(Widget request, Widget new, 
		       ArgList args, Cardinal *narg);
static void Realize(Widget w, XtValueMask *valueMask, 
		    XSetWindowAttributes *attrs);
static void Redisplay(Widget w, XEvent *event, Region region);
static void Resize(Widget w);
static void Destroy(Widget w);
static Boolean SetValues(Widget gcurrent, Widget grequest, Widget gnew,
			 ArgList args, Cardinal *narg);


/* Actions and translations
 */
static void Magnify(Widget w, XEvent *event, String *param, Cardinal *np);
static void QuitPuff(Widget w, XEvent *event, String *param, Cardinal *np);
static void ToggleBox(Widget w, XEvent *event, String *param, Cardinal *np);

static XtActionsRec actions[] = {
    {"magnify",		Magnify},
    {"quitpuff",	QuitPuff},
    {"togglebox",	ToggleBox},
};

static char translations[] =
"\
<Key>1:		magnify(2)	\n\
<Key>2:		magnify(4)	\n\
<Key>3:		magnify(6)	\n\
<Key>4:		magnify(8)	\n\
<Key>5:		magnify(10)	\n\
<Key>6:		magnify(12)	\n\
<Key>7:		magnify(14)	\n\
<Key>8:		magnify(16)	\n\
<Key>9:		magnify(18)	\n\
<Key>0:		magnify(20)	\n\
<Key>-:		magnify(25)	\n\
<Key>=:		magnify(30)	\n\
<Key>b:		togglebox()	\n\
<Key>B:		togglebox()	\n\
Ctrl Shift<Key>Escape:	quitpuff()	\n\
";


/* Static widget class defintion.
 * No special geometry management done.
 */
PuffClassRec puffClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Puff",
    /* widget_size		*/	sizeof(PuffRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* puff fields */
    /* empty			*/	0
  }
};

WidgetClass puffWidgetClass = (WidgetClass) &puffClassRec;


/* The following eleven routines handle the job of magnifying
 * the screen.
 */

static void FindImageInfo(PuffWidget pw)
{
    /* At initialization time, determine if we can use the
     * shared memory extension.  Shared pixmaps are best of all.
     */
    int		majorv, minorv;
    Bool	sharedPixmaps;
    
    pw->puff.useShmPixmap = 0;
    pw->puff.useShmImage = 0;
    
#ifdef MITSHM
    /* Make sure we are on the same machine as the server,
     * in addition to whether the extension exists.
     */
    if ((   MATCH(DisplayString(XtDisplay(pw)), ":0")
	 || MATCH(DisplayString(XtDisplay(pw)), "unix:0"))
	&& XShmQueryVersion(XtDisplay(pw), &majorv, &minorv, &sharedPixmaps)) {
	
	if (sharedPixmaps) {
	    /* Only use shared pixmaps if we can directly
	     * access the pixels in the map.  We have code for
	     * one bit and one byte pixels.
	     */
	    XPixmapFormatValues	*xpfv;
	    int			nformat;
	    int			i;
	    
	    xpfv = XListPixmapFormats(XtDisplay(pw), &nformat);
	    
	    for (i = 0; i < nformat; i++)
		if (xpfv[i].depth == pw->core.depth  &&
		    (xpfv[i].bits_per_pixel == 1  ||
		     xpfv[i].bits_per_pixel == 8)) {
		    pw->puff.useShmPixmap = 1;
		    pw->puff.pixBits = xpfv[i].bits_per_pixel;
		}
	    XFree((char *) xpfv);
	}

	if (!pw->puff.useShmPixmap)
	    pw->puff.useShmImage = 1;	/* Better than nothing... */
    }
#endif
}


static void SetupBitTable(PuffWidget pw)
{
    /* For one bit deep images, create a table to expand
     * eight bits at a time.  For each combination of eight
     * bits (256) create a string of magx bytes to expand them.
     * A new bit table needs to be generated when magx changes.
     */
    char	**bitTable = pw->puff.bitTable;
    int		magx = pw->puff.magx;
    int		i, j, k;
    
    for (i = 0; i < 256; i++) {
	bitTable[i] = CallocType(char, magx);
	MemCheck(bitTable[i]);
	
	if (pw->puff.srcImage->bitmap_bit_order == MSBFirst)
	    for (j = 0; j < magx; j++) {
		bitTable[i][j] = 0;
		for (k = 0; k < 8; k++)
		    if (i & (1 << (7 - (j * 8 + k) / magx)))
			bitTable[i][j] |= 1 << (7 - k);
	    }
	else
	    for (j = 0; j < magx; j++) {
		bitTable[i][j] = 0;
		for (k = 0; k < 8; k++)
		    if (i & (1 << (j * 8 + k) / magx))
			bitTable[i][j] |= 1 << k;
	    }
    }
}


static void FreeBitTable(PuffWidget pw)
{
    /* Free the bit table storage.
     */
    int		i;

    for (i = 0; i < 256; i++) {
	Free(pw->puff.bitTable[i]);
	pw->puff.bitTable[i] = NULL;
    }
}


static void SetupSource(PuffWidget pw)
{
    /* Allocate the source image and shared memory, if needed.
     * A new source image needs to be generated when the source
     * region changes size, and that changes when the magnification
     * changes.
     */
    int		width = pw->puff.srcWidth;
    int		height = pw->puff.srcHeight;
    int		depth = pw->core.depth;

    if (pw->puff.useShmPixmap || pw->puff.useShmImage) {

#ifdef MITSHM
	/* Create the source image first to get bytes_per_line.
	 * Then create the shared memory segment and patch up image.
	 */
	pw->puff.srcImage = 
	    XShmCreateImage(XtDisplay(pw), NULL, depth, ZPixmap, 
			    NULL, &pw->puff.srcShm, width, height);

	pw->puff.srcShm.shmid = 
	    shmget(IPC_PRIVATE, pw->puff.srcImage->bytes_per_line * height,
		   IPC_CREAT|0777);
	Verify(pw->puff.srcShm.shmid >= 0, "shmget failed");

	pw->puff.srcShm.shmaddr = (char *) shmat(pw->puff.srcShm.shmid, 0, 0);
	Verify(pw->puff.srcShm.shmaddr != ((char *) -1), "shmat failed");

	pw->puff.srcImage->data = pw->puff.srcShm.shmaddr;
	pw->puff.srcShm.readOnly = False;

	/* Tell the server about this memory.
	 */
	XShmAttach(XtDisplay(pw), &pw->puff.srcShm);

	/* Create a pixmap using the shared memory segment if we can.
	 */
	if (pw->puff.useShmPixmap)
	    pw->puff.srcPix = 
		XShmCreatePixmap(XtDisplay(pw), XtWindow(pw), 
				 pw->puff.srcShm.shmaddr, &pw->puff.srcShm, 
				 width, height, depth);
#endif

    } else
	/* Create an empty source image for imformation.
	 * Data will be filled in when the image is really grabbed.
	 */
	pw->puff.srcImage = 
	    XCreateImage(XtDisplay(pw), NULL, depth, ZPixmap, 0, 
			 NULL, width, height, BitmapPad(XtDisplay(pw)), 0);
    
    /* See if direct pixel access is possible.  We have code for
     * one and eight bit pixels.
     */
    if (!pw->puff.useShmPixmap)
	pw->puff.pixBits = pw->puff.srcImage->bits_per_pixel;
    pw->puff.directAccess = pw->puff.pixBits == 1 || pw->puff.pixBits == 8;
    
    /* If this is a monochrome image, set up the bit table for
     * faster expansion.
     */
    if (pw->puff.pixBits == 1)
	SetupBitTable(pw);
}


static void SetupDestination(PuffWidget pw)
{
    /* Allocate a destination image and shared memory.
     * A new destination image needs to be generated when
     * the widget changes sizes.
     *
     * The source image should be ready to go at this point.
     *
     * Make sure the destination image width is large enough
     * to accomodate overrun by the expansion code.
     */
    int		width = RoundUp(pw->core.width, pw->puff.srcWidth);
    int		height = pw->core.height;
    int		depth = pw->core.depth;
    
    if (pw->puff.srcImage->bits_per_pixel == 1)
	width = RoundUp(pw->core.width, pw->puff.srcWidth * 8);
    
    if (pw->puff.useShmPixmap || pw->puff.useShmImage) {

#ifdef MITSHM
	/* Create the image first to get bytes_per_line.
	 * Then create the shared memory segment and patch up image.
	 */
	pw->puff.dstImage = 
	    XShmCreateImage(XtDisplay(pw), NULL, depth,ZPixmap, 
			    NULL, &pw->puff.dstShm, width, height);
	
	pw->puff.dstShm.shmid = 
	    shmget(IPC_PRIVATE, pw->puff.dstImage->bytes_per_line * height,
		   IPC_CREAT|0777);
	Verify(pw->puff.dstShm.shmid >= 0, "shmget failed");
	
	pw->puff.dstShm.shmaddr = (char *) shmat(pw->puff.dstShm.shmid, 0, 0);
	Verify(pw->puff.dstShm.shmaddr != ((char *) -1), "shmat failed");
	
	pw->puff.dstImage->data = pw->puff.dstShm.shmaddr;
	pw->puff.dstShm.readOnly = False;
	
	/* Tell the server about this memory.
	 */
	XShmAttach(XtDisplay(pw), &pw->puff.dstShm);

	/* Create a pixmap using the shared memory segment if we can.
	 */
	if (pw->puff.useShmPixmap)
	    pw->puff.dstPix = 
		XShmCreatePixmap(XtDisplay(pw), XtWindow(pw), 
				 pw->puff.dstShm.shmaddr, &pw->puff.dstShm, 
				 width, height, depth);
#endif

    } else {
	/* Unlike the source, allocate the destination mmeory,
	 * since the image will be manufactured from the source.
	 */
	pw->puff.dstImage = 
	    XCreateImage(XtDisplay(pw), NULL, depth, ZPixmap, 0, NULL, 
			 width, height, BitmapPad(XtDisplay(pw)), 0);
	
	pw->puff.dstImage->data = 
	    CallocType(char, pw->puff.dstImage->bytes_per_line * height);
	MemCheck(pw->puff.dstImage->data);
    }
}


static void FreeSource(PuffWidget pw)
{
    /* Free up the source image, shared memory.
     * IMPORTANT: detatch from shared memory segment.	
     */
#ifdef MITSHM
    if (pw->puff.useShmPixmap)
	XFreePixmap(XtDisplay(pw), pw->puff.srcPix);
    if (pw->puff.useShmImage || pw->puff.useShmPixmap)
	XShmDetach(XtDisplay(pw), &pw->puff.srcShm);
#endif

    XDestroyImage(pw->puff.srcImage);

#ifdef MITSHM
    if (pw->puff.useShmImage || pw->puff.useShmPixmap) {
	(void) shmdt(pw->puff.srcShm.shmaddr);
	(void) shmctl(pw->puff.srcShm.shmid, IPC_RMID, 0);
    }
#endif
    
    if (pw->puff.pixBits == 1)
	FreeBitTable(pw);
}

static void FreeDestination(PuffWidget pw)
{
    /* Free up the destination image, shared memory.
     * IMPORTANT: detatch from shared memory segment.	
     */
#ifdef MITSHM
    if (pw->puff.useShmPixmap)
	XFreePixmap(XtDisplay(pw), pw->puff.dstPix);
    if (pw->puff.useShmImage || pw->puff.useShmPixmap)
	XShmDetach(XtDisplay(pw), &pw->puff.dstShm);
#endif

    XDestroyImage(pw->puff.dstImage);

#ifdef MITSHM
    if (pw->puff.useShmImage || pw->puff.useShmPixmap) {
	(void) shmdt(pw->puff.dstShm.shmaddr);
	(void) shmctl(pw->puff.dstShm.shmid, IPC_RMID, 0);
    }
#endif
}


static void PuffPixmap(PuffWidget pw)
{
    /* Time critical, magnification routine.
     * Proceed row by row through the source image.
     * For every source row, expand the first destination row.
     * Expand bytes directly if possible, otherwise use Get/PutPixel.
     * Then copy destination rows to replicate in the y direction.
     */
    char	*dstBase = pw->puff.dstImage->data;
    char	*srcBase = pw->puff.srcImage->data;
    int		srcRowBytes = pw->puff.srcImage->bytes_per_line;
    int		dstRowBytes = pw->puff.dstImage->bytes_per_line;
    int		srcWidth = pw->puff.srcWidth;
    int		i, j, k, l;
    
    if (pw->puff.directAccess  &&  pw->puff.pixBits == 1)
	srcWidth = RoundUp(srcWidth, 8) / 8;
    
    for (j = 0; j < pw->puff.srcHeight; j++) {
	char	*src = srcBase + srcRowBytes * j;
	char	*dst = dstBase + dstRowBytes * j * pw->puff.magy;
	
	k = 0;
	if (pw->puff.directAccess)
	    if (pw->puff.pixBits == 1)
		for (i = 0; i < srcWidth; i++)
		    for (l = 0; l < pw->puff.magx; l++)
			dst[k++] = pw->puff.bitTable[(unsigned char)src[i]][l];
	    else
		for (i = 0; i < srcWidth; i++)
		    for (l = 0; l < pw->puff.magx; l++)
			dst[k++] = src[i];
	else
	    for (i = 0; i < srcWidth; i++) {
		Pixel	p = XGetPixel(pw->puff.srcImage, i, j);
		for (l = 0; l < pw->puff.magx; l++)
		    XPutPixel(pw->puff.dstImage, k++, j * pw->puff.magy, p);
	    }
	
	k = pw->core.height - j * pw->puff.magy;
	k = MIN(k, pw->puff.magy);
	for (i = 1; i < k; i++)
	    (void) memcpy(dst + i * dstRowBytes, dst, dstRowBytes);
    }
}


static void DrawCursor(PuffWidget pw, int origx, int origy)
{
    /* Invert a box at the cursor point
     */
    int		mx, my;

    mx = (pw->puff.mousex - origx) * pw->puff.magx;
    my = (pw->puff.mousey - origy) * pw->puff.magy;
    if (pw->puff.cursorSize > 0  &&
	mx >= 0  &&  mx < pw->core.width  &&
	my >= 0  &&  my < pw->core.height)
	XFillRectangle(XtDisplay(pw), XtWindow(pw), pw->puff.gcXor, 
		       mx - pw->puff.cursorSize * pw->puff.magx, 
		       my - pw->puff.cursorSize * pw->puff.magy,
		       2 * pw->puff.cursorSize * pw->puff.magx,
		       2 * pw->puff.cursorSize * pw->puff.magy);
}


static void DrawViewBox(PuffWidget pw, int origx, int origy)
{
    /* Invert source view box on root window
     */
    int		cs = (pw->puff.magx + pw->puff.cursorSize) / 2;

    XDrawRectangle(XtDisplay(pw), RootWindowOfScreen(pw->core.screen), 
		   pw->puff.gcXor, 
		   origx - cs, origy - cs,
		   pw->puff.srcWidth + 2 * cs, pw->puff.srcHeight + 2 * cs);
}


static void PuffImage(PuffWidget pw, int showFeedback)
{
    /* Grab pixels from the screen into the source,
     * expand the source into the destination,
     * then write the destination to the widget window.
     */
    int		origx, origy;
    Window	root = RootWindowOfScreen(pw->core.screen);
    Display	*dpy = XtDisplay(pw);

    /* Given the center of the source region, 
     * find the top-left corner.
     */
    origx = pw->puff.srcx + pw->puff.bumpx - pw->puff.srcWidth / 2;
    origx = MAX(origx, 0);
    origx = MIN(origx, WidthOfScreen(pw->core.screen) - pw->puff.srcWidth);

    origy = pw->puff.srcy + pw->puff.bumpy - (3 * pw->puff.srcHeight) / 4;
    origy = MAX(origy, 0);
    origy = MIN(origy, HeightOfScreen(pw->core.screen) - pw->puff.srcHeight);

    /* When showing the view box it is necessary to grab the server,
     * in order to prevent another client messing up the inverted region.
     */
    if (showFeedback  &&  pw->puff.viewBox) {
	XGrabServer(dpy);
	DrawViewBox(pw, origx, origy);
    }

    if (pw->puff.useShmPixmap) {

#ifdef MITSHM
	/* XSync so that we know copy has taken place (there is no
	 * other way to tell besides a GraphicsExpose event)
	 */
	XCopyArea(dpy, root, pw->puff.srcPix, pw->puff.gc, 
		  origx, origy, pw->puff.srcWidth, pw->puff.srcHeight, 0, 0);
	XSync(dpy, False);
	PuffPixmap(pw);
	XCopyArea(dpy, pw->puff.dstPix, XtWindow(pw), pw->puff.gc, 
		  0, 0, pw->core.width, pw->core.height, 0, 0);
#endif

    } else if (pw->puff.useShmImage) {
	
#ifdef MITSHM
	XShmGetImage(dpy, root, pw->puff.srcImage,
		     origx, origy, AllPlanes);
	PuffPixmap(pw);
	XShmPutImage(dpy, XtWindow(pw), pw->puff.gc, 
		     pw->puff.dstImage, 0, 0, 
		     0, 0, pw->core.width, pw->core.height, False);
#endif
	
    } else {
	
	XDestroyImage(pw->puff.srcImage);
	pw->puff.srcImage = XGetImage(dpy, root, origx, origy, 
				      pw->puff.srcWidth, pw->puff.srcHeight,
				      AllPlanes, ZPixmap);
	PuffPixmap(pw);
	XPutImage(dpy, XtWindow(pw), pw->puff.gc, pw->puff.dstImage, 
		  0, 0, 0, 0, pw->core.width, pw->core.height);
    }
    
    DrawCursor(pw, origx, origy);

    /* Get rid of the view box and ungrab the server.
     */
    if (showFeedback  &&  pw->puff.viewBox) {
	DrawViewBox(pw, origx, origy);
	XUngrabServer(dpy);
    }
}


static int QueryMouse(PuffWidget pw)
{
    /* Find out where the mouse is.
     * Yes, I know, round trip to the server.  Most pointer
     * activity though is reported via root motion events.
     */
    Window	rwin = RootWindowOfScreen(pw->core.screen);
    Window	cwin;
    int		rootx, rooty;
    int		winx, winy;
    unsigned int mask;
    int		flash;

    (void) XQueryPointer(XtDisplay(pw), rwin,
			 &rwin, &cwin, &rootx, &rooty, &winx, &winy, &mask);
    
    flash = mask || (pw->puff.mousex != rootx) || (pw->puff.mousey != rooty);

    pw->puff.srcWindow = rwin;
    pw->puff.srcXoff = pw->puff.srcYoff = 0;
    pw->puff.mousex = pw->puff.srcx = rootx;
    pw->puff.mousey = pw->puff.srcy = rooty;
    
    return flash;
}


/* ARGSUSED */
static void TimerCB(XtPointer closure, XtIntervalId *id)
{
    /* Consumate a call to PuffSetLocation.
     */
    PuffWidget	pw = (PuffWidget) closure;
    
    PuffImage(pw, TRUE);
    pw->puff.updatePending = FALSE;
}


/* ARGSUSED */
static void PeriodicCB(XtPointer closure, XtIntervalId *id)
{
    /* Update the widget every so often, regardless of activity.
     */
    PuffWidget	pw = (PuffWidget) closure;
    int		flashBox = FALSE;
    
    /* If an update is pending to TimerCB, forget puffing here.
     */
    if (!pw->puff.updatePending) {
	if (pw->puff.lookAtMouse)
	    flashBox = QueryMouse(pw);	/* Keep tracking the mouse */
	PuffImage(pw, flashBox);
    }

    /* Same Bat Time, same Bat Channel!
     */
    if (pw->puff.delayMax)
	XtAppAddTimeOut(XtWidgetToApplicationContext((Widget) pw), 
			pw->puff.delayMax, PeriodicCB, (XtPointer) pw);
}


int PuffCatchError(Widget w, XErrorEvent *ee)
{
    PuffWidget		pw = (PuffWidget) w;

    if (ee->request_code == X_TranslateCoords  &&
	ee->error_code == BadWindow) {
	pw->puff.srcXoff = pw->puff.srcYoff = 0;
	pw->puff.srcWindow = RootWindowOfScreen(pw->core.screen);
	return 1;
    }
    return 0;
}


void PuffSetLocation(Widget w, Window win, int x, int y, int isMouse)
{
    /* Tell the Puff widget to look somewhere.
     * x and y are reported relative to window win.
     * If isMouse is TRUE, this is mouse activity.
     */
    PuffWidget		pw = (PuffWidget) w;
    Window		root = RootWindowOfScreen(pw->core.screen);
    Window		cwin;
    
    /* Cancel arrow key bumping if there is mouse motion.
     */
    if (isMouse)
	pw->puff.bumpx = pw->puff.bumpy = 0;
    pw->puff.lookAtMouse = isMouse;
    
    /* If this is graphics activity and an update is already
     * pending, ignore this location.  This keeps puff
     * from jumping around when there is mouse motion that is also
     * generating graphics activity, like text cursor highlighting.
     */
    if ((!isMouse && pw->puff.updatePending)  || 
	(pw->puff.bumpx || pw->puff.bumpy))
	return;

    /* If this is a new window, find out the offset of the window
     * on the root window.
     */
    if (pw->puff.srcWindow != win) {
	pw->puff.srcWindow = win;
	if (win == root)
	    pw->puff.srcXoff = pw->puff.srcYoff = 0;
	else
	    XTranslateCoordinates(XtDisplay(pw), win, root, 0, 0, 
				  &pw->puff.srcXoff, &pw->puff.srcYoff, &cwin);
    }
    
    /* Set the new location.
     */
    pw->puff.srcx = x + pw->puff.srcXoff;
    pw->puff.srcy = y + pw->puff.srcYoff;
    if (isMouse) {
	pw->puff.mousex = x + pw->puff.srcXoff;
	pw->puff.mousey = y + pw->puff.srcYoff;
    }

    /* Schedule an update if one isn't already pending.
     * Don't update here, because a flood of PuffSetLocations
     * may be happening.  Update only at set intervals (delayMin).
     */
    if (!pw->puff.updatePending) {
	pw->puff.updatePending = TRUE;
	XtAppAddTimeOut(XtWidgetToApplicationContext(w), pw->puff.delayMin,
			TimerCB, (XtPointer) pw);
    }
}


void PuffBump(Widget w, int direction)
{
    /* Move the region of focus around, usually in response to 
     * arrow keys.  Note that when the source region is bumped,
     * PuffSetLocation ignores graphics activity.  This is because
     * the arrow key grabs often generate focus in-out events,
     * which generates graphics, which defeats the bump attempt.
     */
    PuffWidget		pw = (PuffWidget) w;

    pw->puff.lookAtMouse = FALSE;

    switch (direction) {
      case PUFF_UP:
	pw->puff.bumpy -= pw->puff.srcHeight / 4;
	break;
      case PUFF_DOWN:
	pw->puff.bumpy += pw->puff.srcHeight / 4;
	break;
      case PUFF_RIGHT:
	pw->puff.bumpx += pw->puff.srcWidth / 4;
	break;
      case PUFF_LEFT:
	pw->puff.bumpx -= pw->puff.srcWidth / 4;
	break;
      case PUFF_RETURN:
	pw->puff.bumpx = pw->puff.bumpy = 0;
	break;
    }

    /* Schedule an update if one isn't already pending.
     */
    if (!pw->puff.updatePending) {
	pw->puff.updatePending = TRUE;
	XtAppAddTimeOut(XtWidgetToApplicationContext(w), pw->puff.delayMin,
			TimerCB, (XtPointer) pw);
    }
}


static void CalcSourceSize(PuffWidget pw)
{
    /* Make sure that sourceSize * mag > destinationSize
     */
    pw->puff.srcWidth  = (pw->core.width  + pw->puff.magx - 1) / pw->puff.magx;
    pw->puff.srcHeight = (pw->core.height + pw->puff.magy - 1) / pw->puff.magy;
}


static void SetLineWidth(PuffWidget pw)
{
    XSetLineAttributes(XtDisplay(pw), pw->puff.gcXor, 
		       pw->puff.cursorSize + pw->puff.magx - 1,
		       LineSolid, CapButt, JoinMiter);
}


/* ARGSUSED */
static void Initialize(Widget request, Widget new, 
		       ArgList args, Cardinal *narg)
{
    /* Widget initialize method
     */
    PuffWidget	pw = (PuffWidget) new;
    XGCValues	gcv;

    pw->puff.srcWindow = 0;
    pw->puff.srcXoff = 0;
    pw->puff.srcYoff = 0;
    pw->puff.bumpx = 0;
    pw->puff.bumpy = 0;
    pw->puff.mousex = 0;
    pw->puff.mousey = 0;
    pw->puff.updatePending = 0;
    pw->puff.lookAtMouse = 1;
    
    pw->puff.magx = MAX(1, pw->puff.magx);
    pw->puff.magy = MAX(1, pw->puff.magy);
    
    /* Set IncludeInferiors so we can grab all screen pixels using
     * the root window.
     */
    gcv.graphics_exposures = False;
    gcv.subwindow_mode = IncludeInferiors;
    pw->puff.gc = XtGetGC(new, GCGraphicsExposures | GCSubwindowMode, &gcv);

    /* Make sure black inverts to white, and visa versa.
     */
    gcv.foreground = (BlackPixelOfScreen(pw->core.screen) ^
		      WhitePixelOfScreen(pw->core.screen));
    if (pw->core.depth == 1)
	gcv.function = GXinvert;
    else
	gcv.function = GXxor;
    gcv.line_width = pw->puff.cursorSize + pw->puff.magx - 1;
    pw->puff.gcXor = XtGetGC(new, GCGraphicsExposures | GCSubwindowMode |
			     GCFunction | GCForeground | GCLineWidth, &gcv);

    CalcSourceSize(pw);
    FindImageInfo(pw);
}


static void Realize(Widget w, XtValueMask *valueMask, 
		    XSetWindowAttributes *attrs)
{
    /* Widget realize method
     */
    PuffWidget	pw = (PuffWidget) w;

    XtCreateWindow(w, InputOutput, (Visual *) CopyFromParent,
		   *valueMask, attrs);
    
    /* Now that window exists, setup images.
     */
    CalcSourceSize(pw);
    SetupSource(pw);
    SetupDestination(pw);
    
    if (pw->puff.delayMax)
	XtAppAddTimeOut(XtWidgetToApplicationContext(w),
			pw->puff.delayMax, PeriodicCB, (XtPointer) pw);
}


void PuffReleaseShm(Widget w)
{
    /* If you can't destroy the Puff widget, at least call
     * this routine before dying.  It is important to
     * release the shared memory segments.
     */
    PuffWidget	pw = (PuffWidget) w;

#ifdef MITSHM
    if (pw->puff.useShmImage || pw->puff.useShmPixmap) {
	(void) shmdt(pw->puff.srcShm.shmaddr);
	(void) shmctl(pw->puff.srcShm.shmid, IPC_RMID, 0);
	(void) shmdt(pw->puff.dstShm.shmaddr);
	(void) shmctl(pw->puff.dstShm.shmid, IPC_RMID, 0);
    }
#endif
}


static void Destroy(Widget w)
{
    /* Widget destroy method
     */
    PuffWidget	pw = (PuffWidget) w;

    XtDestroyGC(pw->puff.gc);
    if (pw->puff.srcImage) {
	FreeSource(pw);
	FreeDestination(pw);
    }
}


/* ARGSUSED */
static void Redisplay(Widget w, XEvent *event, Region region)
{
    /* Widget expose method
     */
    PuffImage((PuffWidget) w, FALSE);
}


static void Resize(Widget w)
{
    /* Widget resize method
     */
    PuffWidget	pw = (PuffWidget) w;
    
    if (!XtIsRealized(w))
	return;

    FreeSource(pw);
    FreeDestination(pw);

    CalcSourceSize(pw);
    SetupSource(pw);
    SetupDestination(pw);
}


/* ARGSUSED */
static Boolean SetValues(Widget gcurrent, Widget grequest, Widget w,
			 ArgList args, Cardinal *narg)
{
    PuffWidget	current = (PuffWidget) gcurrent;
    PuffWidget	pw = (PuffWidget) w;
    Boolean	redisplay = FALSE;
    
    if (pw->puff.magx != current->puff.magx  ||
	pw->puff.magy != current->puff.magy) {
	
	pw->puff.magx = MAX(1,  pw->puff.magx);
	pw->puff.magy = MAX(1,  pw->puff.magy);
	
	/* The source region needs to be recalculated
	 * when the magnification changes.
	 */
	SetLineWidth(pw);
	FreeSource(pw);
	CalcSourceSize(pw);
	SetupSource(pw);
	redisplay = TRUE;
    }
    
    if (pw->puff.srcx != current->puff.srcx  ||
	pw->puff.srcy != current->puff.srcy) {
	pw->puff.bumpx = pw->puff.bumpy = 0;
	PuffSetLocation(w, RootWindowOfScreen(pw->core.screen),
			pw->puff.srcx, pw->puff.srcy, FALSE);
	redisplay = TRUE;
    }
    
    if (pw->puff.cursorSize != current->puff.cursorSize)
	redisplay = TRUE;

    return redisplay;
}


/*ARGSUSED*/
static void Magnify(Widget w, XEvent *event, String *param, Cardinal *np)
{
    /* magnify action
     */
    PuffWidget	pw = (PuffWidget) w;
    
    if (*np != 1) {
	XtAppWarning(XtWidgetToApplicationContext(w), 
		     "The magnify action takes exactly one argument.");
	XBell(XtDisplay(w), 0);
	return;
    }
    
    pw->puff.magx = pw->puff.magy = atoi(param[0]);
    
    /* The source region needs to be recalculated
     * when the magnification changes.
     */
    SetLineWidth(pw);
    FreeSource(pw);
    CalcSourceSize(pw);
    SetupSource(pw);
    
    PuffImage(pw, FALSE);
}


/*ARGSUSED*/
static void QuitPuff(Widget w, XEvent *event, String *param, Cardinal *np)
{
    /* quitpuff action
     */
    XtUnrealizeWidget(w);
}

/*ARGSUSED*/
static void ToggleBox(Widget w, XEvent *event, String *param, Cardinal *np)
{
    /* togglebox action
     */
    PuffWidget	pw = (PuffWidget) w;
    
    pw->puff.viewBox = !pw->puff.viewBox;
}

