/*
 * $Id: $
 * $Date: $
 *
 * Author:	Kenneth Chin-Purcell, AHPCRC
 *		Copyright @ 1992, Minnesota Supercomputer Center, Inc.
 *
 * Purpose:	magnifier
 *
 * RESTRICTED RIGHTS LEGEND
 *
 * Use, duplication, or disclosure of this software and its documentation
 * by the Government is subject to restrictions as set forth in subdivision
 * { (b) (3) (ii) } of the Rights in Technical Data and Computer Software
 * clause at 52.227-7013.
 *
 * $Log:$
 */

#ifndef _PuffP_h
#define _PuffP_h

#include "Puff.h"
#include <X11/CoreP.h>

#ifdef MITSHM
/* Use a local XShm.h header which has function prototypes.
 * X11R5 may fix this.
 */
#include <sys/ipc.h>
#include <sys/shm.h>
#include "XShm.h"
#endif

typedef struct {
    int			dummy;
} PuffClassPart;

typedef struct _PuffClassRec {
    CoreClassPart	core_class;
    PuffClassPart	puff_class;
} PuffClassRec;

extern PuffClassRec puffClassRec;

typedef struct {
    /* resources */
    int			magx, magy;	/* mag factor for each direction */
    int			srcx, srcy;	/* center of source area */
    int			delayMin;	/* minimum delay, milliseconds */
    int			delayMax;	/* maximum delay, milliseconds */
    int			cursorSize;	/* base size of cursor block */
    int			viewBox;	/* indicate view box */

    /* private */
    int			srcWidth;	/* source width */
    int			srcHeight;	/* source height */
    Window		srcWindow;	/* window ID of last set location */
    int			srcXoff;	/* window x offset on root */
    int			srcYoff;	/* window y offset on root */
    int			mousex, mousey; /* where to place cursor block */
    int			bumpx, bumpy;	/* view offset */
    GC			gc;		/* GCSubwindowMode = true */
    GC			gcXor;		/* GCSubwindowMode = true */
    int			pixBits;	/* bits per pixel */
    int			directAccess;	/* OK to directly access image */
    int			useShmImage;	/* shared memory image extension */
    int			useShmPixmap;	/* shared pixmap extension */
    XImage		*srcImage;	/* source pixels */
    XImage		*dstImage;	/* destination pixels */
#ifdef MITSHM
    XShmSegmentInfo	srcShm;		/* shared memory for source */
    XShmSegmentInfo	dstShm;		/* shared memory for destination */
    Pixmap		srcPix;		/* shared memory source pixmap */
    Pixmap		dstPix;		/* shared memory destination pixmap */
#endif
    char		*bitTable[256];	/* table for mono expansion */
    int			lookAtMouse;	/* focus on mouse (vs. graphics) */
    int			updatePending;	/* window update already pending */
} PuffPart;

typedef struct _PuffRec {
    CorePart		core;
    PuffPart		puff;
} PuffRec;

#endif /* _PuffP_h */
